import functools
from collections import deque
from copy import copy

from pandaSuit.common.mappings.reversible import *
from pandaSuit.common.unwind import Unwind
from pandaSuit.common.constant.decorators import UNWIND_LIST


def extract_function_name(function_reference: str) -> str:
    return function_reference.split(".")[1].split(" at")[0]


def reversible(func):
    """Allow for reversing an 'in place' operation on pandaSuit object"""
    @functools.wraps(func)
    def wrapper_reverse(*args, **kwargs):
        function_name = extract_function_name(func.__repr__())
        intermediate_reverse_function = INTERMEDIATE_REVERSE_MAPPING.get(function_name)
        intermediate_reverse_args = INTERMEDIATE_REVERSE_ARGS.get(function_name)(kwargs)
        reverse_args = REVERSE_ARGS.get(function_name)(kwargs)
        if intermediate_reverse_function is not None:
            reverse_args.update({
                ARGUMENT_MAPPING.get(function_name):
                    copy(args[0].__getattribute__(intermediate_reverse_function)(**intermediate_reverse_args))
            })
        reverse_function = REVERSE_MAPPING.get(function_name)
        args[0].__setattr__(UNWIND_LIST, args[0].__getattribute__(UNWIND_LIST) + deque([Unwind(reverse_function, reverse_args)]))
        func(*args, **kwargs)
    return wrapper_reverse



import tkinter

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from pandas import Series

from pandaSuit.df import DF
from pandaSuit.plot.dashboard.tile import Tile
from pandaSuit.plot.plot import Plot


class Dashboard:
    def __init__(self,
                 root: tkinter.Tk = None,
                 rows: int = 1,
                 columns: int = 1,
                 layout: DF = None,
                 title: str = "",
                 background_color: str = "white"):
        self.layout = layout if layout is not None else DF([[None for _ in range(columns)] for _ in range(rows)])
        self.title = title
        self.background_color = background_color
        self.dashboard = root if root is not None else tkinter.Tk()
        self.dashboard.config(bg=self.background_color)
        try:
            self.dashboard.title(self.title)
        except AttributeError:
            pass
        self._shown = False

    def add_plot(self, plot: Plot, row: int = None, column: int = None) -> None:
        if row is not None and column is not None:
            if self.layout.select(row=row, column=column) is None:
                self.layout.update(row=row, column=column, to=Tile(plot))
            else:
                raise Exception(f"There is already a Dashboard Tile placed at position row={row} column={column}")
        else:
            row, column = self._next_available_position()
            self.add_plot(plot, row, column)

    def update_plot(self, row: int, column: int, to: Plot) -> None:
        try:
            self.layout.update(row=row, column=column, to=Tile(to))
        except IndexError:
            raise Exception(f"Dashboard position ({row}, {column}) specified does not exist.")

    def display(self, pop_out: bool = True) -> None:
        if self._shown:
            rows, columns = self.layout._df.shape
            self = Dashboard(rows=rows, columns=columns, layout=self.layout, title=self.title, background_color=self.background_color)
            self.display()
        else:
            row_count = 0
            for row in self.layout.rows:
                column_count = 0
                for tile in row.to_list():
                    if tile is not None:
                        FigureCanvasTkAgg(tile.figure, master=self.dashboard).get_tk_widget().grid(row=row_count, column=column_count)
                    column_count += 1
                row_count += 1
            self._shown = True
            if pop_out:
                self.dashboard.mainloop()
                self.grab_set()

    def add_row(self, columns: int = None) -> None:
        pass

    def add_column(self, rows: int = None) -> None:
        pass

    def grab_set(self):
        self.dashboard.grab_set()

    # Private methods
    def _next_available_position(self) -> tuple:
        row_count = 0
        for row in self.layout.rows:
            column_count = 0
            for tile in list(row):
                if tile is None:
                    return row_count, column_count
                column_count += 1
            row_count += 1
        self._augment_underlying_table()
        return self._next_available_position()

    def _augment_underlying_table(self) -> None:
        if self.layout.column_count > self.layout.row_count:
            self.layout.append(row=Series([None] * self.layout.column_count), in_place=True)
        else:
            self.layout.append(column=Series([None] * self.layout.row_count), in_place=True)



from pandaSuit.common.constant.decorators import UPDATE


def intermediate_update_args(kwargs):
    return {
        "column": kwargs.get("column")
    }


def update_args(kwargs):
    return {
        "column": kwargs.get("column"),
        "to": kwargs.get("to")
    }


REVERSE_MAPPING = {
    UPDATE: "update"
}

REVERSE_ARGS = {
    UPDATE: update_args
}

INTERMEDIATE_REVERSE_MAPPING = {
    UPDATE: "select"
}

INTERMEDIATE_REVERSE_ARGS = {
    UPDATE: intermediate_update_args
}

ARGUMENT_MAPPING = {
    UPDATE: "to"
}



from math import exp
from random import choice

from sklearn.linear_model import LogisticRegression
from pandas import Series, DataFrame


class LogisticModel:

    def __init__(self, dependent: Series, independent: Series or DataFrame, intercept: bool = True):
        self.dependent = dependent
        self.independent = independent
        self.include_intercept = intercept
        self.model = self._fit()

    def predict(self, input: Series or dict or int or float, practical: bool = False) -> float:
        return self._practical_prediction(input) if practical else self._theoretical_prediction(input)

    def _fit(self) -> LogisticRegression:
        y = self.dependent.to_list()
        if isinstance(self.independent, Series):
            x = self.independent.to_numpy().reshape(-1, 1)
        else:
            x = self.independent
        return LogisticRegression(fit_intercept=self.include_intercept).fit(X=x, y=y)

    @property
    def intercept(self) -> float:
        return self.model.intercept_[0]

    @property
    def betas(self) -> list:
        return list(self.model.coef_.tolist()[0])

    # Private prediction methods
    def _theoretical_prediction(self, input: Series or dict) -> bool:
        if isinstance(input, Series):
            return int(list(self.model.predict(input.to_numpy().reshape(1, -1)))[0]) == 1
        else:
            return int(list(self.model.predict(Series(input).to_numpy().reshape(1, -1)))[0]) == 1

    def _practical_prediction(self, input: Series or dict or int or float) -> float:  # todo
        if isinstance(input, dict):
            for predictor_name, predictor_value in input.items():
                input[predictor_name] = predictor_value + (choice([-1, 1]) * self.independent.standard_deviation(column=predictor_name))
            return self._theoretical_prediction(input)
        elif isinstance(input, Series):
            return self._practical_prediction(input.to_dict())
        elif isinstance(input, (float, int)):
            return self._theoretical_prediction(input + self.independent.std() * choice([-1, 1]))
        else:
            raise Exception("Must supply Series, dict, int or float to .predict() method")



from sklearn.tree import DecisionTreeClassifier
from pandas import Series, DataFrame


class ClassificationTree:
    def __init__(self, dependent: Series, independent: Series or DataFrame):
        self.dependent = dependent
        self.independent = independent
        self.model = self._fit()

    def _fit(self) -> DecisionTreeClassifier:
        y = self.dependent.to_list()
        if isinstance(self.independent, Series):
            x = self.independent.to_numpy().reshape(-1, 1)
        else:
            x = self.independent
        return DecisionTreeClassifier(random_state=0).fit(X=x, y=y)

    def predict(self, input: Series or dict or int or float):
        if isinstance(input, (Series, dict)):
            return self.model.predict(DataFrame([input]).values)[0]
        else:
            return self.model.predict(input)[0]



from abc import ABC, abstractmethod


class Plot(ABC):
    def __init__(self, figure, axes):
        self.figure = figure
        self.axes = axes

    @abstractmethod
    def display(self) -> None:
        return None

    @abstractmethod
    def create_plot(self) -> None:
        return None



from __future__ import annotations

from collections import deque
from random import randint
from copy import copy

import pandas

from pandaSuit.common.unwind import Unwind
from pandaSuit.common.util.list_operations import index_dictionary, create_index_list
from pandaSuit.stats.linear import LinearModel
from pandaSuit.stats.logistic import LogisticModel
from pandaSuit.common.constant.date_constants import DATE_GROUPINGS
from pandaSuit.common.decorators import reversible


class DF:
    def __init__(self, data=None):
        if data is not None:
            self._df = pandas.DataFrame(data)
        else:
            self._df = pandas.DataFrame()
        self._unwind = deque()

    def select(self,
               row: list or int or str = None,
               column: list or int or str = None,
               pandas_return_type: bool = True) -> pandas.DataFrame or pandas.Series or DF:
        if row is None:
            if self._names_supplied(column):
                result = self._df[column]
            else:
                result = self._df.iloc[:, column]
        else:
            if column is None:
                if self._names_supplied(row):
                    result = self._df.loc[row]
                else:
                    result = self._df.iloc[row]
            else:
                if self._names_supplied(row) and self._names_supplied(column):
                    result = self._df.loc[row, column]
                else:
                    if self._names_supplied(row):
                        result = self._df.loc[row].iloc[:, column]
                    else:
                        result = self._df.iloc[row, column]
        if pandas_return_type:
            return result
        else:
            return DF(result)

    def where(self, column_name: str, some_value: object, pandas_return_type: bool = True) -> pandas.DataFrame:
        if isinstance(some_value, str):
            result = self._df[self._df[column_name].str.contains(some_value, na=False)]
        else:
            result = self._df.loc[self._df[column_name] == some_value]
        return result if pandas_return_type else DF(result)

    def where_not(self, column_name: str, some_value: object, pandas_return_type: bool = True) -> pandas.DataFrame:
        if isinstance(some_value, str):
            result = self._df[~self._df[column_name].isin([some_value])]
        else:
            result = self._df.loc[self._df[column_name] != some_value]
        return result if pandas_return_type else DF(result)

    def random_row(self) -> pandas.DataFrame:
        return self._df.iloc[randint(0, self._df.shape[0] - 1)]

    def regress(self, y: str or int, x: list or str or int, logit: bool = False) -> LinearModel or LogisticModel:
        if logit:
            return LogisticModel(dependent=self.select(column=y), independent=self.select(column=x))
        else:
            return LinearModel(dependent=self.select(column=y), independent=self.select(column=x))

    def where_null(self, column: str, pandas_return_type: bool = True) -> DF or pandas.DataFrame:
        result = self._df[self._df[column].isnull()]
        return result if pandas_return_type else DF(result)

    def where_not_null(self, column: str, pandas_return_type: bool = True) -> DF or pandas.DataFrame:
        result = self._df[self._df[column].notna()]
        return result if pandas_return_type else DF(result)

    def group_by(self, column: int or str = None, row: int or str = None, date_grouping: str = None) -> dict:
        """
        Returns a dictionary object that groups on a Row/Column, using the grouping values as keys, pointing to Table objects containing the Row(s)/Column(s) that contain the key value.
        :param column: Column to group on
        :param row: Row to group on
        :param date_grouping: type of date grouping (e.g. "day", "month", "year")
        :return: Dictionary containing values grouped by (keys) and items belonging to that grouping (values).
        """
        if date_grouping is None:
            return {name: self.select(column=indexes, pandas_return_type=False)
                    if row is not None else self.select(row=indexes, pandas_return_type=False)
                    for name, indexes in index_dictionary(
                    (self.select(row=row, pandas_return_type=True) if row is not None
                     else self.select(column=column, pandas_return_type=True)).values).items()}
        else:
            grouping = DATE_GROUPINGS.get(date_grouping)
            if grouping is None:
                raise Exception(f"Invalid date grouping type \"{date_grouping}\"")
            if column is None:
                raise Exception("Cannot group on a Row of dates")
            date_group_by_object = self._df.groupby(pandas.to_datetime(self.select(column=column)).dt.strftime(grouping))
            return {date_key: DF(date_group_by_object.get_group(date_key)) for date_key in list(date_group_by_object.groups.keys())}

    def sum_product(self, *columns: int or str) -> int or float:
        product_column = pandas.Series([1]*self.row_count)
        for column in columns:
            product_column *= self.select(column=column)
        return product_column.sum()

    @reversible
    def update(self, row: int or str = None, column: int or str = None, to: object = None, in_place: bool = True) -> DF or None:
        if in_place:
            if column is not None:
                if isinstance(column, str):
                    self._df.loc[create_index_list(self.row_count), column] = to
                else:
                    self._df.iloc[create_index_list(self.row_count), column] = to
            elif row is not None:
                if isinstance(row, str):
                    pass
                else:
                    pass
            else:
                raise Exception("Please supply a row or column to update.")
        else:
            _df = copy(self)
            _df.update(row=row, column=column, to=to, in_place=True)
            return _df

    def append(self, row: pandas.Series = None, column: pandas.Series = None, in_place: bool = True) -> DF or None:
        if row is not None and column is None:
            if in_place:
                self._append_row(row, in_place)
            else:
                return self._append_row(row, in_place)
        elif row is None and column is not None:
            if in_place:
                self._append_column(column, in_place)
            else:
                return self._append_column(column, in_place)
        elif row is not None and column is not None:
            if len(row) > len(column):
                if in_place:
                    self._append_column(column, in_place)
                    self._append_row(row, in_place)
                else:
                    return DF(copy(self._df))._append_column(column, in_place)._append_row(row, in_place)
            else:
                if in_place:
                    self._append_row(row, in_place)
                    self._append_column(column, in_place)
                else:
                    return DF(copy(self._df))._append_row(row, in_place)._append_column(column, in_place)
        else:
            raise Exception("row or column parameter must be set")

    def undo(self) -> None:
        """
        Reverts the most recent change to the Table instance.
        """
        unwind_object: Unwind = self._unwind.pop()
        self.__getattribute__(unwind_object.function)(**unwind_object.args[0])

    def _append_row(self, row: pandas.Series, in_place: bool) -> DF or None:
        if in_place:
            self._df.append(row, ignore_index=True)
        else:
            _df = copy(self._df)
            _df.append(row, ignore_index=True)
            return DF(_df)

    def _append_column(self, column: pandas.Series, in_place: bool) -> DF or None:
        if in_place:
            self._df.insert(self.column_count, column.name, column, True)
        else:
            _df = copy(self._df)
            _df.insert(self.column_count, column.name, column, True)
            return DF(_df)

    # def undo(self):
    #     """
    #     Reverts the most recent change to the Table instance.
    #     """
    #     try:
    #         unwind_object = self._unwind.pop()
    #         try:
    #             if isinstance(unwind_object, Unwind):
    #                 if len(unwind_object.args) > 0:
    #                     unwind_object.function(unwind_object.args)
    #                 else:
    #                     unwind_object.function()
    #             else:
    #                 for unwind_step in reversed(unwind_object):
    #                     if len(unwind_step.args) > 0:
    #                         unwind_step.function(unwind_step.args)
    #                     else:
    #                         unwind_step.function()
    #             del self._change_log[-1]
    #         except Exception as e:
    #             raise Exception(f"Error occurred when attempting to undo step: {self._change_log[-1]}", e)
    #     except IndexError:
    #         pass

    @staticmethod
    def _names_supplied(selector: int or str or list) -> bool:
        if isinstance(selector, list):
            return isinstance(selector[0], str)
        else:
            return isinstance(selector, str)

    @property
    def is_empty(self) -> bool:
        return self._df.empty

    @property
    def rows(self) -> list:
        return [pandas.Series(row[1]) for row in self._df.iterrows()]

    @property
    def column_count(self) -> int:
        return len(self._df.columns)

    @property
    def row_count(self) -> int:
        return len(self._df)

    def __setattr__(self, name, value):
        try:
            super(DF, self).__setattr__(name, value)
        except AttributeError:  # don't immediately raise AttributeError
            self._df.__setattr__(name, value)  # instead, invoke setter on underlying pandas DataFrame

    def __getattribute__(self, name):
        try:
            return super(DF, self).__getattribute__(name)
        except AttributeError:  # don't immediately raise AttributeError
            return self._df.__getattribute__(name)  # instead, invoke getter on underlying pandas DataFrame



import pathlib

from setuptools import setup, find_packages


setup(
    author='Anthony Raimondo',
    author_email='anthonyraimondo7@gmail.com',
    description='pandas extension for DataFrame manipulation, statistics, and plotting',
    install_requires=["pandas", "scikit-learn"],
    license='MIT',
    long_description=(pathlib.Path(__file__).resolve().parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    name='pandaSuit',
    package_dir={'': 'src/main/python'},
    packages=find_packages(where='src/main/python', exclude='test'),
    url='https://github.com/AnthonyRaimondo/pandaSuit',
    version='1.1.4'
)



from __future__ import annotations

import matplotlib.pyplot as plt

from pandaSuit.plot.plot import Plot


class PiePlot(Plot):
    def __init__(self, slices: dict, **kwargs):
        self._shown = False
        self.slices = slices
        self.figure, self.axes = plt.subplots()
        self.title = kwargs.get('title')
        self.data_labels = kwargs.get("data_labels")
        self.edge_color = kwargs.get("edge_color")
        super().__init__(self.figure, self. axes)

    def display(self) -> None:
        if self._shown:
            self = PiePlot(self.slices)
            self.display()
        else:
            self.create_plot()
            self._shown = True
            plt.show()

    def create_plot(self) -> None:
        self.axes.pie(list(self.slices.values()),
                      labels=list(self.slices.keys()),
                      wedgeprops={'edgecolor': self.edge_color if self.edge_color else 'black'},
                      autopct='%1.1f%%')
        self._add_chart_features()

    def _add_chart_features(self) -> None:
        if self.title is not None:
            self.axes.set_title(self.title)
        # if self.data_labels is not None:
        #     self.axes.set_autopct('%1.1f%%')
